#!/usr/bin/python3

import sys, os.path, argparse, gzip, shutil

import pycurl
from io import BytesIO

from apt_pkg import uri_to_filename, version_compare
from aptsources.sourceslist import SourceEntry
from debian_bundle import deb822


def read_curl(cache_dir, url):
    path = get_path_curl(cache_dir, url)
    with open(path) as f: return f.read()


def zcat_curl(cache_dir, url):
    path = get_path_curl(cache_dir, url)
    with gzip.open(path, 'r') as f: return f.read()


def overwrite_cache(bytesio, cache_filename):
    with open(cache_filename, 'wb') as f: f.write(bytesio.getbuffer())


def get_file_path(url):
    FILE0 = 'file:'
    FILE2 = 'file://'
    FILE3 = 'file:///'

    if not url.startswith(FILE0): return None
    if url.startswith(FILE3): return url[len(FILE2):]
    if url.startswith(FILE2): sys.exit('Invalid url: ' + url)
    return url[len(FILE0):]


def get_path_curl(cache_dir, url):
    file_path = get_file_path(url)
    if file_path is not None: return file_path

    CURL = pycurl.Curl()
    # CURL.setopt(CURL.VERBOSE, 1) # debug
    CURL.setopt(CURL.URL, url)

    filename = os.path.join(cache_dir, uri_to_filename(url))

    if os.path.isfile(filename):
        m_time = os.path.getmtime(filename)
        CURL.setopt(CURL.TIMEVALUE, int(m_time))
        CURL.setopt(CURL.TIMECONDITION, CURL.TIMECONDITION_IFMODSINCE)

    b = BytesIO()
    CURL.setopt(CURL.WRITEDATA, b)
    CURL.perform()
    code = CURL.getinfo(CURL.RESPONSE_CODE)
    CURL.close()

    if code != 213 and code != 304:
        # file modified
        overwrite_cache(b, filename)

    return filename


def list_files(url):
    file_path = get_file_path(url)
    if file_path is not None:
        return [d for d in os.listdir(file_path) if os.path.isdir(os.path.join(file_path, d))]

    CURL = pycurl.Curl()
    # CURL.setopt(CURL.VERBOSE, 1) # debug
    CURL.setopt(CURL.URL, url)

    b = BytesIO()
    CURL.setopt(CURL.WRITEDATA, b)
    CURL.perform()
    code = CURL.getinfo(CURL.RESPONSE_CODE)
    CURL.close()

    index = b.getvalue().decode("utf-8")
    if code == 200:
        rows = index.split('<tr>')
        index = []
        for r in rows:
            pos = r.find('alt="[DIR]"')
            if pos == -1: continue
            pos = r.find('href="', pos) + 6
            epos = r.find('/"', pos)
            index.append(r[pos:epos])
        return index
    elif code == 226:
        index = index.strip().split('\n')
        index = [i.split() for i in index]
        index = [i[-1] for i in index if i[-2] != '->']
        return index
    else:
        sys.exit('Error ' + code + ' accessing ' + url)


def get_dists_uri(entry, component, archpath):
    return entry.uri + '/dists/' + entry.dist + '/' + component + '/' + archpath


def update_cache(cache, p, entry, component, md5):
    name = p['Package']
    p.entry = entry
    p.entry_component = component
    p.md5 = md5

    if not name in cache:
        cache[name] = p
        return

    version = p['Version']
    old_version = cache[name]['Version']
    if version_compare(old_version, version) < 0: cache[name] = p


def download_bin_cache(entry, cache, cache_dir, arch):
    archs = entry.architectures
    if not archs: archs = [arch]
    cache_paths = ['binary-' + a for a in archs]

    for path in cache_paths:
        for c in entry.comps:
            cfc = zcat_curl(cache_dir, get_dists_uri(entry, c, path) + "/Packages.gz")
            for par in deb822.Deb822.iter_paragraphs(cfc):
                p = deb822.Packages(par)
                update_cache(cache, p, entry, c, p['MD5Sum'])


def download_src_cache(entry, cache, cache_dir, arch):
    cache_name = 'Sources'

    for c in entry.comps:
        cfc = zcat_curl(cache_dir, get_dists_uri(entry, c, 'source') + "/Sources.gz")
        for par in deb822.Deb822.iter_paragraphs(cfc):
            p = deb822.Sources(par)
            md5 = p['Files']
            md5 = ''.join([l['MD5Sum'] for l in md5])
            update_cache(cache, p, entry, 'source', md5)


def get_release(cache_dir, canonic_url, codename):
    rfc = read_curl(cache_dir, canonic_url + "/dists/" + codename + "/Release")
    return deb822.Release(rfc)


def get_cache(sources_list_dir, cache_dir, arch):
    sources_list_path = os.path.abspath(os.path.join(sources_list_dir, 'sources.list'))

    src_cache = {}
    bin_cache = {}
    first_entry = True

    with open(sources_list_path) as f:
        for line in f:
            entry = SourceEntry(line)
            if entry.disabled: continue

            if first_entry:
                release = get_release(cache_dir, entry.uri, entry.dist)
                first_entry = False

            if entry.type == 'deb':
                download_bin_cache(entry, bin_cache, cache_dir, arch)
            else:
                download_src_cache(entry, src_cache, cache_dir, arch)

    return src_cache, bin_cache, release


def get_codename(url, codename):
    if url[0] == '/': url = 'file://' + url
    pos = url.find('/dists/')

    if pos != -1:
        codename = url[pos + 7:]
        return url[:pos], codename

    if codename is not None: return url, codename

    l = list_files(url + '/dists/')
    codename = l[0]
    return url, codename


def get_component(name):
    pos = name.find('/')
    return name[:pos]


def prepare_source_list(sources_list_dir, url, codename):
    sources_list_path = os.path.abspath(os.path.join(sources_list_dir, 'sources.list'))
    os.makedirs(sources_list_dir, exist_ok = True)

    if os.path.isfile(url):
        shutil.copyfile(url, sources_list_path)
        return

    canonic_url, codename = get_codename(url, codename)
    release = get_release(sources_list_dir, canonic_url, codename)

    deb_components = set()
    src_components = set()
    for r in release['MD5Sum']:
        name = r['name']
        if name.endswith('/Packages.gz'): deb_components.add(get_component(name))
        elif name.endswith('/Sources.gz'): src_components.add(get_component(name))

    arch = release['Architectures'].replace(' ', ',')

    with open(sources_list_path, 'w') as f:
        if deb_components: print('deb [arch=' + arch + '] ', canonic_url, codename, ' '.join(deb_components), file = f)
        if src_components: print('deb-src', canonic_url, codename, ' '.join(src_components), file = f)

